from typing import Optional, Tuple

from vbeam.apodization.window import Window
from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util import ensure_2d_point
from vbeam.util.geometry.v2 import Line, distance


def rtb_apodization(
    point: np.ndarray,
    array_left: np.ndarray,
    array_right: np.ndarray,
    focus_point: np.ndarray,
    minimum_aperture: float,
    window: Optional[Window] = None,
):
    """Focused apodization used for retrospective transmit beamforming (RTB).

    A focused apodization has two main parts:
    - An hourglass shape constructed by the area between two lines drawn from either
      side of the array and passing through the focus point (line_left and line_right).
    - A line with a thickness given by minimum_aperture that passes through the focus point,
      to account for the fact that the hourglass apodization is infinitely narrow at the
      focus point.

    A point that lies both outside of the hourglass and outside of the thickness of the
    line is weighted by 0.
    """
    # Ensure that all points are 2D (only 2D is supported)
    point = ensure_2d_point(point)
    array_left = ensure_2d_point(array_left)
    array_right = ensure_2d_point(array_right)
    focus_point = ensure_2d_point(focus_point)

    # Set up geometry
    line_left = Line.passing_through(array_left, focus_point)
    line_right = Line.passing_through(array_right, focus_point)
    line_mid = Line.with_angle(focus_point, (line_left.angle + line_right.angle) / 2)
    line_mid_perpendicular = Line(point, line_mid.normal)

    # Short circuit if no window is given (same as giving a Rectangular window)
    if window is None:
        return np.logical_or(
            # Hourglass apodization
            line_left.signed_distance(point) * line_right.signed_distance(point) <= 0,
            # Middle line apodization
            np.abs(line_mid.signed_distance(point)) < minimum_aperture,
        )

    # Calculate distances
    _, intersection_left = line_left.intersect(line_mid_perpendicular)
    _, intersection_right = line_right.intersect(line_mid_perpendicular)
    distance_left = distance(intersection_left, point)
    total_distance = distance(intersection_left, intersection_right)
    distance_mid = np.abs(line_mid.signed_distance(point))

    # Calculate apodization
    p = distance_left / total_distance  # Window parameter
    valid = line_left.signed_distance(point) * line_right.signed_distance(point) <= 0
    hourglass_apodization = window(np.abs(p - 0.5)) * valid
    mid_line_apodization = window(distance_mid / minimum_aperture)

    # Take the max of the two apodizations
    return np.where(
        hourglass_apodization >= mid_line_apodization,
        hourglass_apodization,
        mid_line_apodization,
    )


@traceable_dataclass((("array_bounds", "minimum_aperture", "window")))
class RTBApodization(Apodization):
    array_bounds: Tuple[float, float]
    minimum_aperture: float = 0.001  # TODO: Calculate this based on F# and wavelength
    window: Optional[Window] = None

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        return rtb_apodization(
            point_position,
            self.array_bounds[0],
            self.array_bounds[1],
            wave_data.source,
            self.minimum_aperture,
            self.window,
        )



from typing import Callable, Literal, Optional, Tuple, Union, overload

import numpy

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.interpolation import FastInterpLinspace
from vbeam.scan.base import CoordinateSystem, Scan
from vbeam.scan.util import parse_axes
from vbeam.util import ensure_positive_index
from vbeam.util.arrays import grid
from vbeam.util.coordinate_systems import as_cartesian


@traceable_dataclass(("azimuths", "elevations", "depths", "apex"))
class SectorScan(Scan):
    azimuths: np.ndarray
    elevations: Optional[np.ndarray]  # May be None for 2D scans
    depths: np.ndarray
    apex: np.ndarray

    def get_points(self, flatten: bool = True) -> np.ndarray:
        polar_axis = self.elevations if self.is_3d else np.array([0.0])
        points = grid(self.azimuths, polar_axis, self.depths, shape=(*self.shape, 3))
        points = as_cartesian(points)
        # Ensure that points and apex are broadcastable
        apex = (
            np.expand_dims(self.apex, axis=tuple(range(1, self.ndim)))
            if self.apex.ndim > 1
            else self.apex
        )
        points = points + apex
        if flatten:
            points = points.reshape((self.num_points, 3))
        return points

    def replace(
        self,
        # "unchanged" means that the axis will not be changed.
        azimuths: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        elevations: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        depths: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        apex: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
    ) -> "SectorScan":
        return SectorScan(
            azimuths=azimuths if azimuths != "unchanged" else self.azimuths,
            elevations=elevations if elevations != "unchanged" else self.elevations,
            depths=depths if depths != "unchanged" else self.depths,
            apex=apex if apex != "unchanged" else self.apex,
        )

    def update(
        self,
        azimuths: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        elevations: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        depths: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        apex: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> "SectorScan":
        return self.replace(
            azimuths(self.azimuths) if azimuths is not None else "unchanged",
            elevations(self.elevations) if elevations is not None else "unchanged",
            depths(self.depths) if depths is not None else "unchanged",
            apex(self.apex) if apex is not None else "unchanged",
        )

    def resize(
        self,
        azimuths: Optional[int] = None,
        elevations: Optional[int] = None,
        depths: Optional[int] = None,
    ) -> "SectorScan":
        if elevations is not None and self.elevations is None:
            raise ValueError(
                "Cannot resize elevations because it is not defined on this scan"
            )
        return self.replace(
            azimuths=(
                np.linspace(self.azimuths[0], self.azimuths[-1], azimuths)
                if azimuths is not None
                else "unchanged"
            ),
            elevations=(
                np.linspace(self.elevations[0], self.elevations[-1], elevations)
                if elevations is not None
                else "unchanged"
            ),
            depths=(
                np.linspace(self.depths[0], self.depths[-1], depths)
                if depths is not None
                else "unchanged"
            ),
        )

    @property
    def shape(self) -> Tuple[int, ...]:
        return (
            (len(self.azimuths), len(self.elevations), len(self.depths))
            if self.elevations is not None
            else (len(self.azimuths), len(self.depths))
        )

    @property
    def bounds(self):
        bounds = []
        for axis in [self.azimuths, self.elevations, self.depths]:
            if axis is not None:
                bounds += [axis[0], axis[-1]]
        return tuple(bounds)

    @property
    def cartesian_bounds(self):
        """Get the bounds of the scan in cartesian coordinates. It is the same as
        bounding box of the scan-converted image."""
        if self.is_3d:
            raise NotImplementedError(
                "Cartesian bounds are not implemented for 3D scans yet.",
                "Please create an issue on Github if this is something you need.",
            )
        min_az, max_az, min_d, max_d = self.bounds
        # We get the bounds by calculating the bound for each edge of the bounding box
        # individually. _right_bound gets the right-most x coordinate of the bounding
        # box and we can get the other sides by rotating the azimuth bounds by 90, 180,
        # and 270 degrees. Because in ultrasound, "straight down" is at 0 degrees, we
        # have to rotate everything by an additional 90 degrees.
        quarter_turn = np.pi / 2
        half_turn = np.pi
        # Return (left, right, top, bottom)
        return (
            -_right_bound(min_az + quarter_turn, max_az + quarter_turn, min_d, max_d),
            _right_bound(min_az - quarter_turn, max_az - quarter_turn, min_d, max_d),
            -_right_bound(min_az + half_turn, max_az + half_turn, min_d, max_d),
            _right_bound(min_az, max_az, min_d, max_d),
        )

    @property
    def coordinate_system(self) -> CoordinateSystem:
        return CoordinateSystem.POLAR

    def __repr__(self):
        return f"SectorScan(<shape={self.shape}>, apex={self.apex})"


def _right_bound(
    min_azimuth: float,
    max_azimuth: float,
    min_depth: float,
    max_depth: float,
) -> float:
    """For the two arcs defined by the azimuth bounds (one for ``min_depth`` and one
    for ``max_depth``), find the right-most x coordinate of the two arcs.

    This will be the right-edge of a bounding box that encompasses the arcs. You can
    get the other edges of the bounding box by rotating the azimuth bounds by 90, 180,
    and 270 degrees.

    See ``docs/tutorials/scan/sector_scan_bounds.ipynb`` for a visualization of the 
    bounding box of the arcs."""
    cos_min, cos_max = np.cos(min_azimuth), np.cos(max_azimuth)
    sin_min, sin_max = np.sin(min_azimuth), np.sin(max_azimuth)

    # Get the maximum x coordinate of the corners of both the inner and outer arc.
    max_corner_x = np.max(
        np.array(
            [
                cos_min * min_depth,  # Inner arc
                cos_max * min_depth,  # Inner arc
                cos_min * max_depth,  # Outer arc
                cos_max * max_depth,  # Outer arc
            ]
        )
    )

    # The right-most part of the arcs may either be ``max_corner_x``, or it may be on
    # the right-most *tangent* of the outer arc. We have to make some additional checks
    # to make this work. The code for this is a bit terse, so just trust the generative
    # unit tests for :attr:`SectorScan.cartesian_bounds` :)
    return np.where(
        (max_azimuth - min_azimuth) < np.pi,
        np.where(np.logical_and(sin_min < 0, sin_max > 0), max_depth, max_corner_x),
        np.where(np.logical_or(sin_min < 0, sin_max > 0), max_depth, max_corner_x),
    )


@overload
def sector_scan(
    azimuths: np.ndarray,
    depths: np.ndarray,
    apex: Union[np.ndarray, float] = 0.0,
) -> SectorScan:
    ...  # 2D scan


@overload
def sector_scan(
    azimuths: np.ndarray,
    elevations: np.ndarray,
    depths: np.ndarray,
    apex: Union[np.ndarray, float] = 0.0,
) -> SectorScan:
    ...  # 3D scan


def sector_scan(*axes: np.ndarray, apex: Union[np.ndarray, float] = 0.0) -> SectorScan:
    "Construct a sector scan. See SectorScan documentation for more details."
    azimuths, elevations, depths = parse_axes(axes)
    return SectorScan(azimuths, elevations, depths, np.array(apex))



from hypothesis import strategies as st

from vbeam.fastmath import numpy as np
from vbeam.scan import sector_scan


def increasing_linspaces(min: float, max: float, num: int = 50):
    "Generate linspaces where start <= stop."
    return (
        st.tuples(st.floats(min, max), st.floats(min, max))
        .filter(lambda args: args[0] <= args[1])
        .map(lambda args: np.linspace(*args, num))
    )


def sector_scans(shape):
    "Generate instances of :class:`~vbeam.scan.SectorScan` with the given shape."
    assert len(shape) in (2, 3), "scan must be 2D or 3D"
    return st.builds(
        sector_scan,
        *[increasing_linspaces(-np.pi, np.pi, size) for size in shape[:-1]],
        increasing_linspaces(0, 2, shape[-1]),
        apex=st.sampled_from([np.array([0.1, 0, 0.2])]),
    )



from abc import ABC, abstractmethod

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


class Window(ABC):
    @abstractmethod
    def __call__(self, ratio: float) -> float:
        ...


@traceable_dataclass()
class NoWindow(Window):
    def __call__(self, ratio: float) -> float:
        return 1.0


@traceable_dataclass()
class Rectangular(Window):
    def __call__(self, ratio: float) -> float:
        return (ratio <= 0.5) * 1.0


@traceable_dataclass(("a0", "a1"))
class Hanning(Window):
    a0: float = 0.5
    a1: float = 0.5

    def __call__(self, ratio: float) -> float:
        return (ratio <= 0.5) * (self.a0 + self.a1 * np.cos(2 * np.pi * ratio))


@traceable_dataclass()
class Hamming(Window):
    def __call__(self, ratio: float) -> float:
        return Hanning(0.53836, 0.46164)(ratio)


@traceable_dataclass(("roll",))
class Tukey(Window):
    roll: float

    def __call__(self, ratio: float) -> float:
        p1 = ratio <= (1 / 2 * (1 - self.roll))
        p2 = ratio > (1 / 2 * (1 - self.roll))
        p3 = (ratio < (1 / 2)) * 0.5
        p4 = 1 + np.cos(2 * np.pi / self.roll * (ratio - self.roll / 2 - 1 / 2))
        return p1 + p2 * p3 * p4


def Tukey25() -> Tukey:
    return Tukey(0.25)


def Tukey50() -> Tukey:
    return Tukey(0.5)


def Tukey75() -> Tukey:
    return Tukey(0.75)


def Tukey80() -> Tukey:
    return Tukey(0.8)


@traceable_dataclass()
class Bartlett(Window):
    def __call__(self, ratio: float) -> float:
        return (0.5 - ratio) * 2



from typing import Callable, Literal, Sequence, Union

from spekk import Spec

from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.util.transformations import *


def get_apodization_values(
    apodization: Apodization,
    sender: ElementGeometry,
    point_position: np.ndarray,
    receiver: ElementGeometry,
    wave_data: WaveData,
    spec: Spec,
    dimensions: Sequence[str],
    sum_fn: Union[
        Literal["sum"], Callable[[np.ndarray, Sequence[int]], np.ndarray]
    ] = "sum",
):
    """Return the apodization values for the dimensions. All other relevant
    dimensions are (by default) summed over.

    If you want the apodization values for each transmit and point, you'd call
    setup.get_apodization_values(["transmits", "points"]). Likewise, if you only
    want the points, you'd call setup.get_apodization_values(["points"]). All other
    dimensions are summed over.

    By default, all dimensions not specified in the list are summed over. You can
    override this by passing in a sum_fn."""
    sum_fn = np.sum if sum_fn == "sum" else sum_fn
    all_dimensions = (
        spec["sender"].dimensions
        | spec["point_position"].dimensions
        | spec["receiver"].dimensions
        | spec["wave_data"].dimensions
    )
    calculate_apodization = compose(
        lambda apodization, *args, **kwargs: apodization(*args, **kwargs),
        *[ForAll(dim) for dim in all_dimensions],
        Apply(sum_fn, [Axis(dim) for dim in all_dimensions - set(dimensions)]),
        # Put the dimensions in the order defined by keep
        Apply(np.transpose, [Axis(dim, keep=True) for dim in dimensions]),
        Wrap(np.jit),  # Make it run faster, if the backend supports it.
    ).build(spec)
    return calculate_apodization(
        apodization=apodization,
        sender=sender,
        point_position=point_position,
        receiver=receiver,
        wave_data=wave_data,
    )



import numpy
from hypothesis import given
from vbeam_test_helpers.generators import sector_scans

from vbeam.scan import SectorScan


def _get_cartesian_bounds_brute_force(scan: SectorScan):
    """Just generate the points from the scan and calculate the minimum and maximum
    x-, y- (if 3D), and z-values."""
    points = scan.get_points()
    points -= scan.apex
    min_x, max_x = points[:, 0].min(), points[:, 0].max()
    min_z, max_z = points[:, 2].min(), points[:, 2].max()
    if scan.is_3d:
        min_y, max_y = points[:, 1].min(), points[:, 1].max()
        return (min_x, max_x, min_y, max_y, min_z, max_z)
    if scan.is_2d:
        return (min_x, max_x, min_z, max_z)


@given(sector_scans([200, 2]))
def test_cartesian_bounds_2D(scan: SectorScan):
    bounds_brute_force = _get_cartesian_bounds_brute_force(scan)
    bounds = scan.cartesian_bounds
    numpy.testing.assert_allclose(bounds_brute_force, bounds, rtol=1e-4, atol=1e-12)


@given(sector_scans([10, 10, 10]))
def test_cartesian_bounds_3D(scan: SectorScan):
    bounds_brute_force = _get_cartesian_bounds_brute_force(scan)
    bounds = scan.cartesian_bounds
    numpy.testing.assert_allclose(bounds_brute_force, bounds, rtol=1e-4, atol=1e-12)



import warnings
from dataclasses import dataclass
from typing import Callable, Dict, Literal, Optional, Sequence, Union

from spekk import Spec, trees
from spekk.util.slicing import IndicesT, slice_data, slice_spec

from vbeam.apodization.util import get_apodization_values
from vbeam.core import SignalForPointData
from vbeam.fastmath import numpy as np
from vbeam.scan import Scan
from vbeam.scan.advanced import ExtraDimsScanMixin
from vbeam.util.transformations import *


@dataclass
class SignalForPointSetup(SignalForPointData):
    spec: Spec
    scan: Optional[Scan] = None

    @property
    def slice(self):
        class Slicer:
            def __getitem__(
                _, slices: Sequence[Union[str, IndicesT]]
            ) -> "SignalForPointSetup":
                sliced_data = slice_data(self.data, self.spec, slices)
                sliced_spec = slice_spec(self.spec, slices)
                return SignalForPointSetup(
                    **sliced_data, scan=self.scan, spec=sliced_spec
                )

        return Slicer()

    def __setattr__(self, name: str, value) -> None:
        # Handle cases where scan and point_position are set at the same time
        # Note: scan always takes precedence
        if name == "point_position" and self.scan is not None:
            raise AttributeError(
                "You may not set point_position when a scan has been defined. Try \
updating the scan instead."
            )
        if name == "scan":
            if self.point_position is not None:
                warnings.warn("point_position will be overwritten by the scan.")

        return super().__setattr__(name, value)

    def __getattribute__(self, name: str):
        scan: Scan = object.__getattribute__(self, "scan")
        if name == "point_position" and scan is not None:
            point_position = object.__getattribute__(self, "point_position")
            if point_position is not None:
                warnings.warn(
                    "Both point_position and scan are set. Scan will be used."
                )
            value = scan.get_points()
        elif name == "spec":
            spec: Spec = object.__getattribute__(self, "spec")
            point_position = object.__getattribute__(self, "point_position")
            if scan is None and point_position is None:
                spec = spec.remove_subtree(["point_position"])
            elif isinstance(scan, ExtraDimsScanMixin):
                spec = spec.at["point_position"].set(scan.flattened_points_dimensions)
            value = spec
        else:
            value = object.__getattribute__(self, name)
        return value

    def size(
        self, dimension: Optional[Union[str, trees.Tree]] = None
    ) -> Union[int, Dict[str, int]]:
        """Return the size of the various dimensions of the data.

        ``dimension`` may be None, a string, or a tree of strings. If it is None, then
        a dict of all the dimensions and their sizes is returned. If it is a string,
        then the size of that dimension is returned. If it is a tree of strings, then
        the leaves of the tree is updated to be the size of those dimensions (assuming
        all leaves are dimensions that exists in the spec).

        See Spec.size for details."""
        data = self.data
        if dimension is None:
            # Returns a dict with all dimensions in the spec as keys and sizes as values
            return self.spec.size(data)
        # dimension may be a tree of dimensions. Update the leaves of the tree to get
        # the size of each dimension.
        return trees.update_leaves(
            dimension,
            lambda x: isinstance(x, str),
            lambda dimension: self.spec.size(data, dimension),
        )

    @property
    def data(self) -> dict:
        kernel_data_fields = SignalForPointData.__dataclass_fields__
        data = {k: getattr(self, k) for k in kernel_data_fields}
        return data

    def copy(self) -> "SignalForPointSetup":
        return SignalForPointSetup(**self.data, spec=self.spec, scan=self.scan)

    def get_apodization_values(
        self,
        dimensions: Sequence[str],
        sum_fn: Union[
            Literal["sum"], Callable[[np.ndarray, Sequence[int]], np.ndarray]
        ] = "sum",
    ):
        """Return the apodization values for the dimensions. All other relevant
        dimensions are (by default) summed over.

        If you want the apodization values for each transmit and point, you'd call
        setup.get_apodization_values(["transmits", "points"]). Likewise, if you only
        want the points, you'd call setup.get_apodization_values(["points"]). All other
        dimensions are summed over.

        By default, all dimensions not specified in the list are summed over. You can
        override this by passing in a sum_fn."""
        return get_apodization_values(
            self.apodization,
            self.sender,
            self.point_position,
            self.receiver,
            self.wave_data,
            self.spec,
            dimensions,
            sum_fn,
        )



from functools import reduce
from operator import mul
from typing import Iterable

from vbeam.core import Apodization
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(("apodizations",))
class CombinedApodization(Apodization):
    apodizations: Iterable[Apodization]

    def __call__(self, *args, **kwargs) -> float:
        """Multiply the result of calling the Apodization objects."""
        values = [apodization(*args, **kwargs) for apodization in self.apodizations]
        return reduce(mul, values, 1.0)


def combine_apodizations(
    *apodizations: Iterable[Apodization],
) -> CombinedApodization:
    """Return a new Apodization object that combines all the given apodizations
    by multiplication."""
    return CombinedApodization(apodizations)
